﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.Security.Policy;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.Bandits.GrammarPolicies {
  public class BoltzmannExplorationPolicy : GrammarPolicy {

    private readonly HashSet<string> done;
    private readonly double beta;

    public BoltzmannExplorationPolicy(IProblem problem, bool useCanonicalRepresentation = false, int beta = 100)
      : base(problem, useCanonicalRepresentation) {
      this.beta = beta;
      this.done = new HashSet<string>();
    }

    public override bool TrySelect(Random random, ReadonlySequence curState, IEnumerable<ReadonlySequence> afterStates, out ReadonlySequence selectedState) {
      // only select states that are not yet done
      afterStates = afterStates.Where(a => !done.Contains(CanonicalState(a.ToString()))).ToArray();
      if (!afterStates.Any()) {
        // fail because all follow states have already been visited => also disable the current state
        done.Add(CanonicalState(curState.ToString()));
        selectedState = null;
        return false;
      }
      // try each state at least once
      if (afterStates.Any(s => GetTries(s) == 0)) {
        selectedState = afterStates.First(s => GetTries(s) == 0);
        return true;
      }

      var w = from state in afterStates
              select Math.Exp(beta * AvgReward(state));

      selectedState = afterStates.SampleProportional(random, w).First();
      return true;
    }

    public override void UpdateReward(IEnumerable<ReadonlySequence> stateTrajectory, double reward) {
      base.UpdateReward(stateTrajectory, reward);
      // the last state could be terminal
      var lastState = stateTrajectory.Last();
      if (lastState.IsTerminal) done.Add(CanonicalState(lastState.ToString()));
    }

    public void Reset() {
      base.Reset();
      done.Clear();
    }
  }
}
