﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Eventing.Reader;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.Bandits.GrammarPolicies {
  public class GreedyPolicy : GrammarPolicy {
    private readonly IProblem problem;

    public GreedyPolicy(IProblem problem, bool useCanonicalRepresentation)
      : base(problem, useCanonicalRepresentation) {
    }

    public override bool TrySelect(Random random, ReadonlySequence curState, IEnumerable<ReadonlySequence> afterStates, out ReadonlySequence selectedState) {
      // always returns the afterstate with the highest value
      double maxV = V(afterStates.First());
      var bestStates = new List<ReadonlySequence>();
      bestStates.Add(afterStates.First());
      foreach (var state in afterStates.Skip(1)) {
        if (maxV < V(state)) {
          maxV = V(state);
          bestStates.Clear();
          bestStates.Add(state);
        } else if (maxV == V(state)) {
          bestStates.Add(state);
        }
      }
      // break ties randomly
      // selectedState = bestStates.SelectRandom(random);
      // break ties using first
      selectedState = bestStates.First();
      return true;
    }

    private double V(ReadonlySequence state) {
      return base.AvgReward(state);
    }
  }
}
