﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.Security.Policy;
using System.Text;
using HeuristicLab.Common;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.Bandits.GrammarPolicies {
  public class RandomNoResamplingPolicy : GrammarPolicy {

    private readonly HashSet<string> done;

    public RandomNoResamplingPolicy(IProblem problem, bool useCanonicalRepresentation)
      : base(problem, useCanonicalRepresentation) {
      this.done = new HashSet<string>();
    }

    public override bool TrySelect(Random random, ReadonlySequence curState, IEnumerable<ReadonlySequence> afterStates, out ReadonlySequence selectedState) {
      // only select states that are not yet done
      afterStates = afterStates.Where(a => !done.Contains(a.ToString())).ToArray();
      if (!afterStates.Any()) {
        // fail because all follow states have already been visited => also disable the current state
        done.Add(CanonicalState(curState.ToString()));
        selectedState = null;
        return false;
      }

      selectedState = afterStates.SelectRandom(random);
      return true;
    }

    public override void UpdateReward(IEnumerable<ReadonlySequence> stateTrajectory, double reward) {
      base.UpdateReward(stateTrajectory, reward);
      // ignore rewards but update the set of visited terminal states

      // the last state could be terminal
      var lastState = stateTrajectory.Last();
      if (lastState.IsTerminal) done.Add(CanonicalState(lastState.ToString()));
    }

    public override void Reset() {
      base.Reset();
      done.Clear();
    }
  }
}
