﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace HeuristicLab.Problems.GrammaticalOptimization {
  // counts the number of times a pair of symbols occurs in a sentence
  public class RoyalPairProblem : IProblem {
    private const string grammarString = @"
G(S):
S -> a | aS | b | bS
";

    private readonly IGrammar grammar;
    public RoyalPairProblem() {
      this.grammar = new Grammar(grammarString);
    }

    public double BestKnownQuality(int maxLen) {
      return maxLen - 1;
    }

    public IGrammar Grammar {
      get { return grammar; }
    }

    private Regex regex = new Regex("(?=ab)|(?=ba)"); // count the number of "ab" and "ba" pairs
    public double Evaluate(string sentence) {
      // sentence must contain only terminal symbols, we are not checking if the sentence is syntactically valid here because it would be too slow!
      Debug.Assert(sentence.Any(c => grammar.IsTerminal(c)));
      return regex.Matches(sentence).Count;
    }

    public string CanonicalRepresentation(string terminalPhrase) {
      return terminalPhrase;
    }
  }
}
